#ifndef _PROCESSOR_H_
#define _PROCESSOR_H_
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/*
 *
 *  Copyright Peter Boyle, University of Glasgow, 2000.
 *  This software is provided for NON-COMMERCIAL use only,
 *  and may not be redistributed without permission.
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

/*Some limits for convenience for multi-dimensional arrays*/
#define MAX_WIDTH 8
#define MAX_CYCLES (2048*4)
#define MAX_ARG 7
#define MAX_REGS 128

/*Define the types of instruction - later called issue groups*/
/*Philosophy is to define all types that are likely*/
/*to map to execution units */

enum EXECUTIONPIPES{
  FMACPIPE = 0,
  FADDPIPE,
  FMULPIPE,
  FMOVPIPE,
  FLODPIPE,
  FSTRPIPE,
  IALUPIPE,
  LOADPIPE,
  STORPIPE,
  BRCHPIPE,
  CACHPIPE,
  DIRECTIVE,
  WEIRDRULES,  
  PRAGMA, /*Allows optimisation hints to be passed to the scheduler*/
  N_PIPE_TYPES
};

/*Bit mask to check if we can issue to a pipe*/
#define FMACMASK (1<<FMACPIPE)
#define FADDMASK (1<<FADDPIPE)
#define FMULMASK (1<<FMULPIPE)
#define FMOVMASK (1<<FMOVPIPE)
#define IALUMASK (1<<IALUPIPE)
#define LOADMASK (1<<LOADPIPE)
#define BRCHMASK (1<<BRCHPIPE)
#define STORMASK (1<<STORPIPE)
#define FLODMASK (1<<FLODPIPE)
#define FSTRMASK (1<<FSTRPIPE)
#define CACHMASK (1<<CACHPIPE)

/*Define the instructions we may want to use*/

enum FMAC_OPS{
  FMADD,FMSUB,FNMADD,FNMSUB,nfmacs
} ;
enum FMOV_OPS{
  FMOV,
  FNEG,
  nfmovs
};
enum FADD_OPS{
  FADD,FSUB,nfadds
};

enum FMUL_OPS{
  FMUL,nfmuls
};

enum LOAD_OPS{
  ILD_EA,
  LOAD_REG_EA,
  LOAD_ADDR,
  LOAD_IMM,
  nloads
};

enum BRANCH_OPS{
  BRANCH_GT,
  BRANCH_EQ,
  BRANCH_LT,
  BRANCH,
  BRANCH_CTR, /*PowerPC specific - try to use as platform independent*/
  /* a paradigm as possible unless there really *is* a performance hit*/
  BRANCH_GE,
  BRANCH_LE,
  BRANCH_RET,
  nbrch
};

enum STORE_OPS{
  SAVE_REG_EA,
  IST_EA,
  nstore
};
enum FLOAD_OPS{
  FLOAD_EA,
  FLOAD_REG_EA,
  nfloads
};
enum FSAVE_OPS{
  FSAVE_EA,
  FSAVE_REG_EA,
  nfsaves
};
enum WEIRD_OPS{ /*Only used if we want to print dependencies */
  /*These are fictitious instructions which denote stall type*/
  /*when the stall is not caused by a single other instruction*/
  /*Only supported on PPC 440 currently                       */
  WARSTALL, LSPIPESPARSE, CACHESTALL, nweirdops
} ;
enum PRAGMA_OPS{ 
  /*Scheduler hints*/
  DCBT_SPACE,  /*Space between DCBT's or equivalent  */
  DCBT_PRE,    /*No L/S use for N-cycles before DCBT */
  DCBT_POST,   /*No L/S use for N-cycles after  DCBT */
  LOAD_LIM,    /*Insert a LS pipe bubble if N loads in a row */
  STORE_LIM,   /*Insert a LS pipe bubble if N stores in a row */
  LS_LIM,      /*Insert a LS pipe bubble if N loads or stores in a row */
  STORE_SPACE,
  STORE_INORDER,   /*Preserve order of stores */
  npragmas
};

/*CACHE CONTROL INSTRUCTIONS - NEED TO THINK ABOUT PRECISION*/
/*AND CACHELINE INFORMATION     */
/*I added the mtctr ppc op here */
/*This is really becoming a 1 argument category with no effect on "architectural" regs*/
/*Should really change the PIPE category name to reflect this*/
enum CACHE_OPS{
  PREF_IMM,
  WRITE_HINT,
  COUNTER_HINT,
  FLUSH,
  ncache
};

enum IALU_OPS{
  IADD,
  ISUB,
  IMUL,
  IOR, /*NB OR RD,RS,RS == MOV RD,RS*/
  CMOVGE,/*NOT PRESENT ON THE POWER PC - use branch with hint*/
  CMOVLE,/*NOT PRESENT ON THE POWER PC - use branch with hint*/
  CMOVGT,/*NOT PRESENT ON THE POWER PC - use branch with hint*/
  CMOVLT,/*NOT PRESENT ON THE POWER PC - use branch with hint*/
  ISUB_PREDICATE, /*Predicate recording subtraction PPC only*/
                  /*Since we count down in loops and subtract in*/
                  /*bounds checking, I think this is the only predicate*/
                  /*Recording instruction needed....*/
  IOR_PREDICATE,  /*Predicate recording or PPC only*/
  IMUL_IMM,       /*Multiply by immediate value - needed to workout offsets*/
                  /*from site no and atom immediate*/
  IADD_IMM,
  nadds
};

/*Assembler Directives (like branch target, align, etc)*/
enum ASM_OPS{
  Target,
  CmovTarget,
  Cache_Align,
  Enter_Routine,
  Exit_Routine,
  Pipe_Flush,
  FNOP,
  NOP,
  PUSH_WINDOW,
  POP_WINDOW,
  LS_BARRIER,
  ndirective
};

/*Define all types of argument to mnemonics*/
/*Presently just Fregs, Iregs, Effective Addresses = SIMM(Ireg), and immediate val SIMM,Label */
/*Condition register - an ireg on alpha, and a predicate on PPC*/
enum ARGUMENTS{
  Fregs, 
  Iregs,
  SIMM,
  Label,
  Pregs, /*Predicate register - unused at present*/
  Cycles /*Cycle count for scheduler pragmas*/
};

enum ARGDIRS{
inarg,
outarg
};

/*Define a register file - should really have size as a parameter but */
/*Easy to sort later as required I think*/
struct regs{
  int regfile ;
  const char *mnemonic[MAX_REGS];
  int allocated[MAX_REGS] ;
  const char *fmt;
};

/*
 * Each execution unit gets a pipeline - slightly
 * clunky when issue width is narrower than the number of units.
 * but then would need to encode issue rules. 
 * Should be alright just doing it this way since we want to keep the FPU
 * busy at all times, so just pretend the combination of other all units 
 * constitutes (a) seperate pipe(s)
 *
 */
struct pipeline{
  int issuegroups; /*Bit mask flagging which kinds of instruction are allowed to issue*/
};

struct instruction{
  char *mnemonic;
  int narg;
  int pipe;
  int code;
  int argtypes[4];  /*Argument types*/
  int inout[4];     /*read or write - use enum names inarg and outarg*/
  int args[4];      /*The values of the arguments */
  struct instruction *staller;
  int latency;      
};

struct instructionset
{
  struct instruction *instructions[N_PIPE_TYPES];
};



/*
 * This is where we say how the different instructions can issue 
 * on a given processor.
 */
struct processor{
  int npipes; /*Number of issues allowed*/
  int id;
  struct pipeline * pipes;
  struct regs * iregs;
  struct regs * fregs;
  struct instructionset *iset;
  int FP_size;
  int I_size;
  int CacheLine;
  int ABI_argbase;
  int StackPointer;
  int ReturnPointer;
  long Isave;
  long Fsave;
  int switchargs[N_PIPE_TYPES];
  int latency[N_PIPE_TYPES];
  int usenops;
  int delayslot;
  int RegWindows;
  int Bias;
};

/*
 * Dependency types 
 */
enum DEPENDENCIES{
  NO_DEPENDENCY=0,
  READ_A_WRITE , /*True dependency - must wait instruction latency*/
  WRITE_A_READ , /*Can be fixed with a temporary reg*/
  BARRIER        /*raising messes up semantics */
                 /*i .e. cant move instructions across a branch etc...*/
};


/*Linked list of instructions that gets reordered by the scheduler*/
struct inst_queue_entry{
  struct instruction *inst;
  struct inst_queue_entry *prev;
  struct inst_queue_entry *next;
};


/*Linked list to contain all registered immediate arguments*/
struct offset {
  int offst;
  const char * name;
  struct offset *next;
};


/*
 * Defines prefetch and write hint streams used by auto managing 
 * stream functions
 */

enum PATTERNS{
LINEAR,
STRIDED,
LOOKUP,
STREAM_IN,
STREAM_OUT,
LOOKUP_MULTI,
POINTER
};

struct stream {
  int inout;   /*Direction*/
  int pointer; /*Register for cur pointer to */
  int basereg; /*Register for base pointer to matrix*/
  int preg;    /*Register for pointer to prefetch from*/
  int atom; /* handle for the immediate value that defines the bytes per site*/
  int pattern; /*Name of the access pattern (previous enum)*/
  int stride;  /*handle for immediate value of stride (equals atom if LINEAR)*/
  int lines;
  int minline;
  int counter;
  int lookupreg; /*A pointer to a look up shift table*/
  int way;       /*Adding multi-way lookup tables*/
  int nway;
};


/*
 * A software rotating register ... hope to also be able to
 * map to IA64 rotating regs and Hitachi VPP reg windows
 */
struct rotating_reg{
  int kind;
  int *regs;
  int last;
  int num;
};

/*
 *
 * Define functions which are common to all processors
 * this is really the user interface or proto-language definition
 * 
 * Though it is stretching things a bit to call this a language
 *
 */

/*Set the processor of choice*/
void set_processor(struct processor *);
void set_abi_caller_save_all(void);

/* Register management routines*/
int defargcount(int narg);
void getarg(int ireg);
int allocate_reg( int kind,const char *name); /*kind = Iregs or Fregs*/
void free_reg( int kind,int num);

struct rotating_reg *create_rotating_reg(int kind,int len,const char *name);
int get_rotating_register(struct rotating_reg *rp);
void reset_rotating_register(struct rotating_reg *rp);

/*Immediate argument management*/
int def_offset(int units, const char * name);
int def_byte_offset(int units, const char * name);
int get_offset( int handle );
const char * get_offset_name( int handle );
struct offset * get_offset_struct( int handle );

/*Loop management routines*/
int get_target_label(void);

/*Generic instruction insertion*/
void make_inst(int PIPE,int OP,...);

void arg_print(struct instruction *i, int j);

/*Stream optimisation set up routine*/

extern int N_Ahead; /*Streaming distance ahead - 3 times around outer loop*/
struct stream * create_stream(int atom,int base_reg,int countreg,
                                                  int dir,int kind,...);
void iterate_stream(struct stream *);
void iterate_table(struct stream *); /*Introduced for share multi stream tables*/
void set_lookup_nxt_pointer (struct stream *mystream,int nxt);

/*Simplified and multi instruction insertion routines*/
void queue_prefetch  (struct stream *mystream);
void queue_write_hint(struct stream *mystream);
void queue_fload     (int dest,int Himm,int basereg);
void queue_fstore    (int src ,int Himm,int basereg);
void queue_iload     (int dest,int Himm,int basereg);
void queue_load_addr (int dest,int Himm,int basereg );
void queue_instruction(struct instruction *inst);
void queue_iadd( int dest, int a, int b );
void queue_iadd_imm( int dest, int a, int himm );
void queue_imul( int dest, int a, int b );
void queue_imul_imm( int dest, int a, int himm );
void queue_fadd( int dest, int a, int b );
void queue_fmul( int dest, int a, int b );
void queue_fsub( int dest, int a, int b );
void queue_fneg(int dest,int src);
void queue_fmov(int dest,int src);
void queue_fmac(int acct,int a,int b); /*d += a*b*/
void queue_fnmac(int acc,int a,int b); /*d -= a*b*/
void queue_fmadd(int dest,int a,int b,int c); /*d = c+ a*b */
void queue_fnmsub(int dest,int a,int b,int c);/*d = c- a*b */
void queue_istore ( int reg, int offhandle, int basereg );

void pragma(int,int);

/*My crude optimiser*/
void schedule_for_proc(void);

/*Output routines*/
void set_human_readable (int h);
void dump_instruction_queue(void);
void dump_regs(int);
void inst_print(struct instruction *i);

/*These routines depend on whether or not you have an FMADD instruction*/
void setup_cmadds(void); 
                 /*
		  * If we have no FMADD pipe we need to allocate temporary
		  *  registers to chain multiplies and adds
                  */
void queue_cmul    (int e,int f,int a,int b,int c,int d);
void queue_cmadd   (int e,int f,int a,int b,int c,int d);
void queue_two_cmuls(int e1,int f1,int a1,int b1,int c1,int d1,
                     int e2,int f2,int a2,int b2,int c2,int d2);
void queue_two_cmadds(int e1,int f1,int a1,int b1,int c1,int d1,
		      int e2,int f2,int a2,int b2,int c2,int d2);
void queue_conjmul (int e,int f,int a,int b,int c,int d);
void queue_conjmadd(int e,int f,int a,int b,int c,int d);
void queue_two_conjmuls(int e1,int f1,int a1,int b1,int c1,int d1,
			int e2,int f2,int a2,int b2,int c2,int d2);
void queue_two_conjmadds(int e1,int f1,int a1,int b1,int c1,int d1,
			 int e2,int f2,int a2,int b2,int c2,int d2);
void queue_bounds_check ( int work, int dst );
void queue_barrier(void);
void queue_mbar(void);
void queue_pref_set_predicate( int counter );
void queue_iload_imm( int dest,int imm);
void queue_three_cmuls(int e1,int f1,int a1,int b1,int c1,int d1,
                       int e2,int f2,int a2,int b2,int c2,int d2,
                       int e3,int f3,int a3,int b3,int c3,int d3);
void queue_three_cmadds(int e1,int f1,int a1,int b1,int c1,int d1,
                        int e2,int f2,int a2,int b2,int c2,int d2,
                        int e3,int f3,int a3,int b3,int c3,int d3);
void queue_three_cmaddtos(int h1,int g1,int e1,int f1,int a1,int b1,int c1,int d1,
                        int h2,int g2,int e2,int f2,int a2,int b2,int c2,int d2,
                        int h3,int g3,int e3,int f3,int a3,int b3,int c3,int d3);
void queue_three_conjmuls(int e1,int f1,int a1,int b1,int c1,int d1,
                          int e2,int f2,int a2,int b2,int c2,int d2,
                          int e3,int f3,int a3,int b3,int c3,int d3);
void queue_three_conjmadds(int e1,int f1,int a1,int b1,int c1,int d1,
                           int e2,int f2,int a2,int b2,int c2,int d2,
                           int e3,int f3,int a3,int b3,int c3,int d3);
void queue_three_conjmaddtos(int h1,int g1,int e1,int f1,int a1,int b1,int c1,int d1,
                           int h2,int g2,int e2,int f2,int a2,int b2,int c2,int d2,
                           int h3,int g3,int e3,int f3,int a3,int b3,int c3,int d3);

/*
 * Save and restore registers as defined by the ABI - so probably define in
 * processor specific c code
 */

int grab_stack ( int );
void free_stack ( void );
void save_regs(void);
void restore_regs(void);

#ifdef alreg 
#error
#endif
#ifdef def_off
#error
#endif

/*Macros for defining registers and immediate arguments*/
void *arralloc(size_t size, int ndim, ...);

#define def_off( A , B ) int A = def_offset ( B, (char *)#A )
#define alreg(A,T)  int A = allocate_reg(T,(char *)#A)

#define reg_array_1d( C, T , d )        register_array_1d C((char *)#C,T,d) 
#define reg_array_2d( C, T , d1,d2 )    register_array_2d C((char *)#C,T,d1,d2) 
#define reg_array_3d( C, T , d1,d2,d3 ) register_array_3d C( (char *) #C,T,d1,d2,d3) 

#define offset_1d( C, d )        offset_array_1d C((char *)#C,d) 
#define offset_2d( C, d1,d2 )    offset_array_2d C((char *)#C,d1,d2) 
#define offset_3d( C, d1,d2,d3 ) offset_array_3d C((char *)#C,d1,d2,d3) 

#define offset_1d_shft( C, d,s )        offset_array_1d C((char *)#C,d,s) 
#define offset_2d_shft( C, d1,d2,s )    offset_array_2d C((char *)#C,d1,d2,s) 
#define offset_3d_shft( C, d1,d2,d3,s ) offset_array_3d C((char *)#C,d1,d2,d3,s) 

/*Define the available processors*/
enum procids {
  ALPHA264,
  ALPHA264_SINGLE,
  ALPHA164,
  ALPHA164_SINGLE,
  ALPHA064,
  ALPHA064_SINGLE,
  PPC440,
  PPC440_SINGLE,
  POWERIII,
  POWERIII_SINGLE,
  USPARCII,
  USPARCII_SINGLE,
  UNKNOWN,
  UNKNOWN_SINGLE,
  nprocids
};
extern char *procargs[nprocids];

/*Processor dependent functions -> use set of function pointers*/
void set_scheduler_weird(void);
struct processor * create_processor(int);
int start_loop(int counter);
void stop_loop(int branchno,int counter);
void directive_print(struct instruction *i);
void queue_cmovge(int cond,int src,int dst);
void cache_print(struct instruction *i);
void check_iterations(int cntreg,int retno);

void set_processor_optarg(char *opt);
void print_eff_addr(struct instruction *i);
int get_I_size(void);
#endif



















